/* network.c
 * 
 *  a collection of general network functions.
 */

#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "config.h"
#include "extern.h"
#include "defines.h"

void start_listener(int sd) {

	int clientlen;
	struct hostent *host;
	struct sockaddr_in clientaddr;
	char *ipaddr;
	int err=0;

	if (debug) print_debug(1, "Waiting for connections on Port %d.\n", config->listenport);
	
	while(1) {


		clientlen = sizeof(clientaddr);

		config->thread->sockfd = accept(sd, (struct sockaddr *) &clientaddr, &clientlen);

		if (debug) print_debug(2, "Accepted incoming connection - processing...\n");

		ipaddr=(char *)strdup(inet_ntoa(clientaddr.sin_addr));
		host=gethostbyaddr(&clientaddr.sin_addr, sizeof(clientaddr.sin_addr), PF_INET);
		if (!host) {
			herror("Error");
			config->hostname=(char *)calloc(sizeof(char), strlen(ipaddr)+3);
			sprintf(config->hostname, "[%s]", ipaddr);
		} else {
			config->hostname=(char *)strdup(host->h_name);
		}
		if (debug) print_debug(3, "Connection from remote host: %s [%s]\n", config->hostname, ipaddr);
		if (sd<0) {
			perror("Fatal: cannot accept connection");
			cleanup();
			exit(1);
		}
		fflush(stderr);
		fflush(stdout);
		if (!fork()) {
			/* child */
			do_connection(ipaddr);
			/* this will never return, however a catch.... */
			exit(-1);
		} else {
			/* parent */
			free(ipaddr);
			close(config->thread->sockfd);
		}
	}
}

int setup_port(void) {

	int connection=0;
	int sockfd=0;

	struct hostent *host;
	struct sockaddr_in servaddr;

	sockfd=socket(PF_INET, SOCK_STREAM, 0);

	if (config->listenaddr) {
		host=gethostbyname(config->listenaddr);
	} else {
		host=gethostbyname("0.0.0.0");
	}

	servaddr.sin_family=host->h_addrtype;
	servaddr.sin_port=htons(config->listenport);
	//memcpy((char *) &servaddr.sin_addr.s_addr, host?host->h_addr_list[0]:INADDR_ANY, host?host->h_length:sizeof(INADDR_ANY));
	memcpy((char *) &servaddr.sin_addr.s_addr, host->h_addr_list[0], host->h_length);

	connection=bind(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr));

	if (connection) {
		perror("Fatal: cannot bind to port");
		cleanup();
		exit(1);
	}

	listen(sockfd,5);
	return sockfd;

}

/* net_connect(): 
 * accepts							returns
 * hostname (char *)
 * port number (int)
 *											socket file descriptor number
 *
 * this function is just to open a socket
 * to a remote host given the hostname and
 * port number.  It returns the fd number of
 * the open socket.
 */
int net_connect(const char *hostname, const int port) {

	int sockfd, connection, ret;
	int error=0;
	struct hostent *ipaddr;
	struct sockaddr_in servaddr;
	struct timeval tv; 
	fd_set fdw, fdr;
	int i;

	if (verbose) printf("Connecting to %s:%i\n", hostname, port);
	if (debug) print_debug(4, "net_connect(): hostname=%s, port=%d\n", hostname, port);
	ipaddr=gethostbyname(hostname);
	if (!ipaddr) {
		for(i=0;i<5;i++) {
			ipaddr=gethostbyname(hostname);
			if (ipaddr) break;
			sleep(1);
		}
		herror("Fatal Error");
		return(0);
	}
	sockfd=socket(PF_INET, SOCK_STREAM, 0);
	fcntl(sockfd, F_SETFL, O_NONBLOCK);

	servaddr.sin_family=ipaddr->h_addrtype;
	servaddr.sin_port=htons(port);
	memcpy((char *) &servaddr.sin_addr.s_addr, ipaddr->h_addr_list[0], ipaddr->h_length);
	connection=connect(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr));

	tv.tv_sec = 30; 
	tv.tv_usec = 0; 

	if (connection) {
		switch (errno) {
			case EAGAIN: /* we can handle this error */
			case EINPROGRESS: 

				FD_ZERO(&fdw);
				FD_SET(sockfd, &fdw);
				fdr = fdw ;
	
				tv.tv_sec = NET_TIMEOUT; 
				tv.tv_usec = 0; 

				if (select(sockfd + 1, &fdr, &fdw, NULL, &tv ) != 0) {
					/* even if socket is writeable, may be an error, need explicit check */
					if (FD_ISSET(sockfd, &fdr) || FD_ISSET(sockfd, &fdw)) {
						int len = sizeof(error);
						if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (char *)&error, &len) == 0) {
							/* solaris getsockopt() will return -1 if an
 								 * error has occurred  otherwise we must
 								 * explicitly check "error"
							 */
							if (error == 0) { /* no errors on connect */
								if (verbose) printf("Connection Established, testing....\n");

								/* set 30 second tiemouts on read/writes */
								tv.tv_sec = NET_DTIMEOUT; 
								tv.tv_usec = 0; 

								setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)); 
								setsockopt(sockfd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv)); 
														
								break; /* let's get out of here ... */
							}
						}
					}
				}
				/* Fall through to default */
			default: 
				if (verbose) fprintf(stderr, "Timed Out connection.\n");
				close(sockfd); /* close socket, and (maybe) try again */
				sockfd=0;
				break;
		}
	}
	return sockfd;
}

int scan_for_dot (char *msg) {

	int t;

	t=strlen(msg);
	msg+=t;
	msg--;

	if (debug) print_debug(5, "scan_for_dot(): %d %d %d %d %d\n", *(msg - 4), *(msg - 3),
									*(msg - 2), *(msg - 1), *msg);
	return ((*(msg - 4) == '\r') && (*(msg - 3) == '\n') && (*(msg - 2) == '.')
									&& (*(msg - 1) == '\r') && (*msg == '\n'));

}
												
void do_thru_connect(int localfd, int remotefd) {

	int num;

	fd_set fds_pair;
	fd_set fds_copy;
	struct timeval tv;
	int fdmax, timeout;
	char *buf;
	
	/* clear string contents and initialise all to null. Good and safe practise */	
	buf=calloc(sizeof(char), NET_BUFFER+1);	

	FD_ZERO(&fds_pair);
	FD_SET(localfd, &fds_pair);
	FD_SET(remotefd, &fds_pair);

	if (localfd > remotefd) {
		fdmax=localfd+1;
	} else {
		fdmax=remotefd+1;
	}
	timeout = 600; /* we set the timeout to fairly long here 'cause we shouldn't
			* _need_ to time it out...
			*/

	while (1) {
		tv.tv_sec = timeout;
		tv.tv_usec = 0;
	
		(void) memcpy(&fds_copy, &fds_pair, sizeof(fds_pair));

		if (!select(fdmax, &fds_copy, NULL, NULL, &tv)) {
			fprintf(stderr, "do_thru_connect(): Timed out connection\n");
			break;
		}

		if (FD_ISSET(localfd, &fds_copy)) {
			if ((num=read(localfd, buf, NET_BUFFER-1)) <= 0 ) {
				if (debug) print_debug(5, "do_thru_connect(): local read returned: %d\n", num);
				break;
			}
			net_nsend(remotefd, buf, num);
		}

		if (FD_ISSET(remotefd, &fds_copy)) {
			if ((num=read(remotefd, buf, NET_BUFFER-1)) <= 0) {
				if (debug) print_debug(5, "do_thru_connect(): local read returned: %d\n", num);
				break;
			}
			net_nsend(localfd, buf, num);
		}
	}
	free(buf);
	if (debug) print_debug(5, "do_thru_connect(): exiting\n");
}
	 
char *net_read(int sockfd) {	

	int num, r;
	char *str;

	fd_set rdfs;
	struct timeval tv;
	int ret, timeout, i;
	char *buf, *ptr;
	
	unsigned char t[1];


	/* clear string contents and initialise all to null. Good and safe practise */	
	str=calloc(sizeof(char), NET_BUFFER+1);	

	FD_ZERO(&rdfs);
	FD_SET(sockfd, &rdfs);

	timeout = 45;
	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	ret=select((sockfd+1), &rdfs, NULL, NULL, &tv);
	// if(debug==5) print_debug(5, "Select() returned: %d\n", ret);
	if (ret) {
		num=read(sockfd, str, NET_BUFFER);

		if (num<0) {
			perror("An Error Occurred");
			num=0;
		}
	}

	if (!num || !ret)	{
		free(str);
		return NULL; /* safely returned something worth while. */	
	}


	if(debug==5) {
		printf("Debug[5]: net_read(%i) ", num);
		for(r=0;r<num;r++) {
			t[0]=*(str+r);
			printf("%hhi ", t[0]);
		}
		printf("\n");
	}

	/* return the damn stuff.  */	
	return str;
}

int net_send(int sockfd, char *str) {
	
	int r;

	if(debug) print_debug(4, "Sending: %s", str);
	if (!send(sockfd, str, strlen(str), 0)) {
		perror("Error sending to network");
		exit(1);
	}

}

int net_nsend(int sockfd, char *str, int len) {
	
	int r;
	unsigned char t[1];

	if(debug==5) {
		printf("Debug[5]: net_nsend() ");
		for(r=0;r<len;r++) {
			t[0]=*(str+r);
			printf("%hhi ", t[0]);
		}
		printf("\n");
	}
	if (!send(sockfd, str, len, 0)) {
		perror("Error sending to network");
		exit(1);
	}

}
char *getip(const char *hostname) {

	int sockfd, connection;
	struct hostent *ipaddr;
	struct sockaddr_in servaddr;
	int i;

	ipaddr=gethostbyname(hostname);
	if (!ipaddr) {
		for(i=0;i<5;i++) {
			ipaddr=gethostbyname(hostname);
			if (ipaddr) break;
			sleep(1);
		}
		if (!ipaddr) {
			fprintf(stderr, "Fatal Error unable to reverse map hostname %s", hostname);
			herror("in DNS");
			return((char *)hostname);
		}
	}
	return ipaddr->h_addr_list[0];
}

